#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include"SimpleImage.h" /* 画像クラス */
#include"BMPIO.h" /* BMP画像の入出力関数 stdlib.hが必要 */

/* 正規化用の関数 */
void Normalize(Image *R,Image *G,Image *B){
  double max,min;
  max= min = R->img[0][0];
  int i,j;
  
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      if(max<R->img[i][j])max=R->img[i][j];
      if(max<G->img[i][j])max=G->img[i][j];
      if(max<B->img[i][j])max=B->img[i][j];
      if(min>R->img[i][j])min=R->img[i][j];
      if(min>G->img[i][j])min=G->img[i][j];
      if(min>B->img[i][j])min=B->img[i][j];
    }
  double range = fabs(max-min);if(range<=0.0)range=1.0;
   
  for(i=0;i<R->sy;i++)
     for(j=0;j<R->sx;j++){
       double val = 255.0*(R->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      R->img[i][j] = val;
      val = 255.0*(G->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      G->img[i][j] = val;
      val = 255.0*(B->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      B->img[i][j] = val;
     }
     
}
/* グレースケールの標準偏差を計算する関数 */
double getStandardDeviation(Image *R,Image *G,Image *B){
  double ave=0.0;
  double sd=0.0;
  int i,j;
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      double gray = (R->img[i][j]+G->img[i][j]+B->img[i][j])/3.0;
      ave += gray;
      sd += gray*gray;
    }
  ave /= ((double)(R->sx*R->sy));
  sd /= ((double)(R->sx*R->sy));
  sd -= ave;
  sd = sqrt(sd);
  return sd;

}


/* BMP画像を読み込んでBilateralフィルタを行いBMP画像として保存する 
コンパイル: g++ -o ColorBilateralFilter ColorBilateralFilter.cxx -lm
実行： ./ColorBilateralFilter 入力BMP 出力BMP 座標空間Gauss重みの標準偏差sigma_s(double) 輝度Gauss重みの標準偏差sigma_t(double) 畳み込みの半径(int)
引数５

注意点：
sigma_sは5.0～30.0ぐらいの値がよい.
畳み込みの半径は2*sigma_sがよいが、大きなsigma_sだと計算が終わらないので、5～15程度.
sigma_tは輝度値(画像全体, or 平坦にしたい領域)の標準偏差を用いる: 0.1～0.5*(標準偏差)ぐらいの値がよい

*/
int main(int argc,char *argv[]){
  
  
  /* 画像クラスImageを用いて入力画像のカラー用にR,G,Bを宣言する */
  Image *R = NULL;
  Image *G = NULL;
  Image *B = NULL;
  

  /* 画像サイズ */
  int sx,sy;

  
  /* BMPIOクラスを宣言しnewする */
  BMPIO *mybmp = new BMPIO();
  /* BMPIOクラスのreadBMPSize()を使って、argv[1]で与えられたファイル名でsx,syに画像サイズを入れる */
  if(mybmp->readBMPSize(&sx,&sy,argv[1])){
    
    /* sx,syでR,G,Bのメモリを確保する(new する) */
    R = new Image(sx,sy);
    G = new Image(sx,sy);
    B = new Image(sx,sy);
    
    /* BMPIOクラスのreadBMP()を使ってargv[1]で与えられるファイル名の画像をR,G,Bに読み込む */
    mybmp->readBMP(R,G,B,argv[1]);
    
  }else{
    delete mybmp;
    printf("Unsupported BMP format !\n");
    return 1;
  }
  
  /* パラメータ 3つ*/
  /* 座標空間Gauss重みの標準偏差をsigma_sへセットする */
  double sigma_s = atof(argv[3]);

  /* 効率化のためGauss関数の中でsigSをかけるだけにするための変数 */
  double sigS = 1.0/(2.0*sigma_s*sigma_s);

  /* 輝度Gauss重みの標準偏差sigma_tへセットする(ただし今回は、sigma_t*(標準偏差)を使う) */
  double sigma_t = atof(argv[4]);   

  /* 輝度値(グレースケール)の画像全体の標準偏差を計算し、standard_deviationへセットする */
  double standard_deviation = getStandardDeviation(R,G,B);
  printf("sigma_t = %lf, standard_deviation = %lf, (sigma_t*standard_deviation) = %lf\n",sigma_t,standard_deviation,(sigma_t*standard_deviation));
  /* 輝度Gauss重みの標準偏差sigma_tへ標準偏差をかける */
  sigma_t *=standard_deviation; 
  
  /* 効率化のためGauss関数の中でsigIをかけるだけにするための変数 */
  double sigI = 1.0/(2.0*sigma_t*sigma_t);

  /* 畳み込みの半径をradiiへセットする */
  int radii = atoi(argv[5]);

  
  /* 出力画像用にImageクラスをoutR,outG,outBとして宣言しsx,syのサイズでメモリを確保する */
  Image *outR = new Image(sx,sy);
  Image *outG = new Image(sx,sy);
  Image *outB = new Image(sx,sy);
  
  

     
  /* Bilateralフィルタここから */
  
  /* 各画素をたどるループ  */  
  int i,j;
  for(i=0;i<sy;i++)
    for(j=0;j<sx;j++){
      /* R,G,Bを足し合わせるための変数を用意し、ゼロを代入する */
      double dR=0.0;
      double dG=0.0;
      double dB=0.0;
      
      int l,m,u,v;
      
     
      double weight_sum=0.0;
      /* (2*radii+1)x(2*radii+1)の局所Window内をたどるループ */  
      for(l=-radii;l<=radii;l++){
	v = l+i;if(v<0)v=0;if(v>=sy)v=sy-1;
	for(m=-radii;m<=radii;m++){
	  u = m+j;if(u<0)u=0;if(u>=sx)u=sx-1;
	  
	  /* カラー勾配を計算 */
	  double difR = R->img[v][u]-R->img[i][j];
	  double difG = G->img[v][u]-G->img[i][j];
	  double difB = B->img[v][u]-B->img[i][j];
	  
	  /* カラー勾配強度(2乗)の計算 
	     注意点：3.0で割らなくてもよいが、グレースケールのときのパラメータと大きさを揃えるため3.0で割る
	   */
	  double grad = (difR*difR+difG*difG+difB*difB)/3.0;
	  /* 空間の距離 */
	  double dx = ((double)(m*m));
	  double dy = ((double)(l*l));
	  /* 重みの宣言 */
	  double weight;
	  /* ここを編集してweightに　重みの計算 G_{sigma_s}(x-u,y-v)*G_{sigma_t}(I(x,y)-I(u,v))　*/



	  

	  /* ここまで */

	  /* 正規化重み付き平均の為にweightの総和を計算 */
	  weight_sum+=weight;
	  /* ここを編集：画素の座標[v][u]のR,G,B値にweightをかけdR,dG,dBへ足し合わせる:重み付き和  */
	  




	  /* ここまで */
	  
	}
      }
      /* エラーチェック */
      if(weight_sum<=0.0)weight_sum=1.0;
      
      
      /* 出力のoutR,outG,outBへnormalized convolutionとして代入　*/
      outR->img[i][j] = dR/weight_sum;
      outG->img[i][j] = dG/weight_sum;
      outB->img[i][j] = dB/weight_sum;
      
      
    }
  
  
  /* Bilateralフィルタ計算ここまで */
  /* outR,outG,outBを0-255へ線形補間で正規化する */
  Normalize(outR,outG,outB);
  
  
  /* BMPIOクラスのsaveBMP()を使ってargv[2]で与えられるファイル名にoutR,outG,outBの画像を保存する */
  mybmp->saveBMP(outR,outG,outB,argv[2]);
  
  
  /* BMPIOクラスのメモリを開放する(deleteする) */
  delete mybmp;
  
  /* R,G,Bのメモリを開放する(deleteする) */
  delete R;
  delete G;
  delete B;
  
  /* outR,outG,outBのメモリを開放する(deleteする) */
  delete outR;
  delete outG;
  delete outB;
 

  return 0;
}
